package org.eclipse.jdt.internal.compiler.ast;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.ast.CastExpression;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.OperatorExpression;
import org.eclipse.jdt.internal.compiler.codegen.BranchLabel;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.UnconditionalFlowInfo;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.ArrayBinding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;

public class BinaryExpression extends OperatorExpression {
  public Expression left;
  
  public Expression right;
  
  public Constant optimizedBooleanConstant;
  
  public BinaryExpression(Expression left, Expression right, int operator) {
    this.left = left;
    this.right = right;
    this.bits |= operator << 6;
    this.sourceStart = left.sourceStart;
    this.sourceEnd = right.sourceEnd;
  }
  
  public BinaryExpression(org.eclipse.jdt.internal.compiler.ast.BinaryExpression expression) {
    this.left = expression.left;
    this.right = expression.right;
    this.bits = expression.bits;
    this.sourceStart = expression.sourceStart;
    this.sourceEnd = expression.sourceEnd;
  }
  
  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo) {
    try {
      if (this.resolvedType.id == 11)
        return (FlowInfo)this.right.analyseCode(
            currentScope, flowContext, 
            (FlowInfo)this.left.analyseCode(currentScope, flowContext, flowInfo).unconditionalInits())
          .unconditionalInits(); 
      this.left.checkNPE(currentScope, flowContext, flowInfo);
      UnconditionalFlowInfo unconditionalFlowInfo = this.left.analyseCode(currentScope, flowContext, flowInfo).unconditionalInits();
      if ((this.bits & 0xFC0) >> 6 != 2)
        flowContext.expireNullCheckedFieldInfo(); 
      this.right.checkNPE(currentScope, flowContext, (FlowInfo)unconditionalFlowInfo);
      unconditionalFlowInfo = this.right.analyseCode(currentScope, flowContext, (FlowInfo)unconditionalFlowInfo).unconditionalInits();
      if ((this.bits & 0xFC0) >> 6 != 2)
        flowContext.expireNullCheckedFieldInfo(); 
      return (FlowInfo)unconditionalFlowInfo;
    } finally {
      flowContext.recordAbruptExit();
    } 
  }
  
  public void computeConstant(BlockScope scope, int leftId, int rightId) {
    if (this.left.constant != Constant.NotAConstant && 
      this.right.constant != Constant.NotAConstant) {
      try {
        this.constant = 
          Constant.computeConstantOperation(
            this.left.constant, 
            leftId, (
            this.bits & 0xFC0) >> 6, 
            this.right.constant, 
            rightId);
      } catch (ArithmeticException arithmeticException) {
        this.constant = Constant.NotAConstant;
      } 
    } else {
      this.constant = Constant.NotAConstant;
      optimizedBooleanConstant(
          leftId, (
          this.bits & 0xFC0) >> 6, 
          rightId);
    } 
  }
  
  public Constant optimizedBooleanConstant() {
    return (this.optimizedBooleanConstant == null) ? this.constant : this.optimizedBooleanConstant;
  }
  
  public void generateCode(BlockScope currentScope, CodeStream codeStream, boolean valueRequired) {
    BranchLabel falseLabel;
    int pc = codeStream.position;
    if (this.constant != Constant.NotAConstant) {
      if (valueRequired)
        codeStream.generateConstant(this.constant, this.implicitConversion); 
      codeStream.recordPositionsFrom(pc, this.sourceStart);
      return;
    } 
    switch ((this.bits & 0xFC0) >> 6) {
      case 14:
        switch (this.bits & 0xF) {
          case 11:
            codeStream.generateStringConcatenationAppend(currentScope, this.left, this.right);
            if (!valueRequired)
              codeStream.pop(); 
            break;
          case 10:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.iadd(); 
            break;
          case 7:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.ladd(); 
            break;
          case 8:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.dadd(); 
            break;
          case 9:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.fadd(); 
            break;
        } 
        break;
      case 13:
        switch (this.bits & 0xF) {
          case 10:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.isub(); 
            break;
          case 7:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.lsub(); 
            break;
          case 8:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.dsub(); 
            break;
          case 9:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.fsub(); 
            break;
        } 
        break;
      case 15:
        switch (this.bits & 0xF) {
          case 10:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.imul(); 
            break;
          case 7:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.lmul(); 
            break;
          case 8:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.dmul(); 
            break;
          case 9:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.fmul(); 
            break;
        } 
        break;
      case 9:
        switch (this.bits & 0xF) {
          case 10:
            this.left.generateCode(currentScope, codeStream, true);
            this.right.generateCode(currentScope, codeStream, true);
            codeStream.idiv();
            if (!valueRequired)
              codeStream.pop(); 
            break;
          case 7:
            this.left.generateCode(currentScope, codeStream, true);
            this.right.generateCode(currentScope, codeStream, true);
            codeStream.ldiv();
            if (!valueRequired)
              codeStream.pop2(); 
            break;
          case 8:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.ddiv(); 
            break;
          case 9:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.fdiv(); 
            break;
        } 
        break;
      case 16:
        switch (this.bits & 0xF) {
          case 10:
            this.left.generateCode(currentScope, codeStream, true);
            this.right.generateCode(currentScope, codeStream, true);
            codeStream.irem();
            if (!valueRequired)
              codeStream.pop(); 
            break;
          case 7:
            this.left.generateCode(currentScope, codeStream, true);
            this.right.generateCode(currentScope, codeStream, true);
            codeStream.lrem();
            if (!valueRequired)
              codeStream.pop2(); 
            break;
          case 8:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.drem(); 
            break;
          case 9:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.frem(); 
            break;
        } 
        break;
      case 2:
        switch (this.bits & 0xF) {
          case 10:
            if (this.left.constant != Constant.NotAConstant && 
              this.left.constant.typeID() == 10 && 
              this.left.constant.intValue() == 0) {
              this.right.generateCode(currentScope, codeStream, false);
              if (valueRequired)
                codeStream.iconst_0(); 
              break;
            } 
            if (this.right.constant != Constant.NotAConstant && 
              this.right.constant.typeID() == 10 && 
              this.right.constant.intValue() == 0) {
              this.left.generateCode(currentScope, codeStream, false);
              if (valueRequired)
                codeStream.iconst_0(); 
              break;
            } 
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.iand(); 
            break;
          case 7:
            if (this.left.constant != Constant.NotAConstant && 
              this.left.constant.typeID() == 7 && 
              this.left.constant.longValue() == 0L) {
              this.right.generateCode(currentScope, codeStream, false);
              if (valueRequired)
                codeStream.lconst_0(); 
              break;
            } 
            if (this.right.constant != Constant.NotAConstant && 
              this.right.constant.typeID() == 7 && 
              this.right.constant.longValue() == 0L) {
              this.left.generateCode(currentScope, codeStream, false);
              if (valueRequired)
                codeStream.lconst_0(); 
              break;
            } 
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.land(); 
            break;
          case 5:
            generateLogicalAnd(currentScope, codeStream, valueRequired);
            break;
        } 
        break;
      case 3:
        switch (this.bits & 0xF) {
          case 10:
            if (this.left.constant != Constant.NotAConstant && 
              this.left.constant.typeID() == 10 && 
              this.left.constant.intValue() == 0) {
              this.right.generateCode(currentScope, codeStream, valueRequired);
              break;
            } 
            if (this.right.constant != Constant.NotAConstant && 
              this.right.constant.typeID() == 10 && 
              this.right.constant.intValue() == 0) {
              this.left.generateCode(currentScope, codeStream, valueRequired);
              break;
            } 
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.ior(); 
            break;
          case 7:
            if (this.left.constant != Constant.NotAConstant && 
              this.left.constant.typeID() == 7 && 
              this.left.constant.longValue() == 0L) {
              this.right.generateCode(currentScope, codeStream, valueRequired);
              break;
            } 
            if (this.right.constant != Constant.NotAConstant && 
              this.right.constant.typeID() == 7 && 
              this.right.constant.longValue() == 0L) {
              this.left.generateCode(currentScope, codeStream, valueRequired);
              break;
            } 
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.lor(); 
            break;
          case 5:
            generateLogicalOr(currentScope, codeStream, valueRequired);
            break;
        } 
        break;
      case 8:
        switch (this.bits & 0xF) {
          case 10:
            if (this.left.constant != Constant.NotAConstant && 
              this.left.constant.typeID() == 10 && 
              this.left.constant.intValue() == 0) {
              this.right.generateCode(currentScope, codeStream, valueRequired);
              break;
            } 
            if (this.right.constant != Constant.NotAConstant && 
              this.right.constant.typeID() == 10 && 
              this.right.constant.intValue() == 0) {
              this.left.generateCode(currentScope, codeStream, valueRequired);
              break;
            } 
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.ixor(); 
            break;
          case 7:
            if (this.left.constant != Constant.NotAConstant && 
              this.left.constant.typeID() == 7 && 
              this.left.constant.longValue() == 0L) {
              this.right.generateCode(currentScope, codeStream, valueRequired);
              break;
            } 
            if (this.right.constant != Constant.NotAConstant && 
              this.right.constant.typeID() == 7 && 
              this.right.constant.longValue() == 0L) {
              this.left.generateCode(currentScope, codeStream, valueRequired);
              break;
            } 
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.lxor(); 
            break;
          case 5:
            generateLogicalXor(currentScope, codeStream, valueRequired);
            break;
        } 
        break;
      case 10:
        switch (this.bits & 0xF) {
          case 10:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.ishl(); 
            break;
          case 7:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.lshl(); 
            break;
        } 
        break;
      case 17:
        switch (this.bits & 0xF) {
          case 10:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.ishr(); 
            break;
          case 7:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.lshr(); 
            break;
        } 
        break;
      case 19:
        switch (this.bits & 0xF) {
          case 10:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.iushr(); 
            break;
          case 7:
            this.left.generateCode(currentScope, codeStream, valueRequired);
            this.right.generateCode(currentScope, codeStream, valueRequired);
            if (valueRequired)
              codeStream.lushr(); 
            break;
        } 
        break;
      case 6:
        generateOptimizedGreaterThan(
            currentScope, 
            codeStream, 
            null, 
            falseLabel = new BranchLabel(codeStream), 
            valueRequired);
        if (valueRequired) {
          codeStream.iconst_1();
          if ((this.bits & 0x10) != 0) {
            codeStream.generateImplicitConversion(this.implicitConversion);
            codeStream.generateReturnBytecode((Expression)this);
            falseLabel.place();
            codeStream.iconst_0();
            break;
          } 
          BranchLabel endLabel;
          codeStream.goto_(endLabel = new BranchLabel(codeStream));
          codeStream.decrStackSize(1);
          falseLabel.place();
          codeStream.iconst_0();
          endLabel.place();
        } 
        break;
      case 7:
        generateOptimizedGreaterThanOrEqual(
            currentScope, 
            codeStream, 
            null, 
            falseLabel = new BranchLabel(codeStream), 
            valueRequired);
        if (valueRequired) {
          codeStream.iconst_1();
          if ((this.bits & 0x10) != 0) {
            codeStream.generateImplicitConversion(this.implicitConversion);
            codeStream.generateReturnBytecode((Expression)this);
            falseLabel.place();
            codeStream.iconst_0();
            break;
          } 
          BranchLabel endLabel;
          codeStream.goto_(endLabel = new BranchLabel(codeStream));
          codeStream.decrStackSize(1);
          falseLabel.place();
          codeStream.iconst_0();
          endLabel.place();
        } 
        break;
      case 4:
        generateOptimizedLessThan(
            currentScope, 
            codeStream, 
            null, 
            falseLabel = new BranchLabel(codeStream), 
            valueRequired);
        if (valueRequired) {
          codeStream.iconst_1();
          if ((this.bits & 0x10) != 0) {
            codeStream.generateImplicitConversion(this.implicitConversion);
            codeStream.generateReturnBytecode((Expression)this);
            falseLabel.place();
            codeStream.iconst_0();
            break;
          } 
          BranchLabel endLabel;
          codeStream.goto_(endLabel = new BranchLabel(codeStream));
          codeStream.decrStackSize(1);
          falseLabel.place();
          codeStream.iconst_0();
          endLabel.place();
        } 
        break;
      case 5:
        generateOptimizedLessThanOrEqual(
            currentScope, 
            codeStream, 
            null, 
            falseLabel = new BranchLabel(codeStream), 
            valueRequired);
        if (valueRequired) {
          codeStream.iconst_1();
          if ((this.bits & 0x10) != 0) {
            codeStream.generateImplicitConversion(this.implicitConversion);
            codeStream.generateReturnBytecode((Expression)this);
            falseLabel.place();
            codeStream.iconst_0();
            break;
          } 
          BranchLabel endLabel;
          codeStream.goto_(endLabel = new BranchLabel(codeStream));
          codeStream.decrStackSize(1);
          falseLabel.place();
          codeStream.iconst_0();
          endLabel.place();
        } 
        break;
    } 
    if (valueRequired)
      codeStream.generateImplicitConversion(this.implicitConversion); 
    codeStream.recordPositionsFrom(pc, this.sourceStart);
  }
  
  public void generateOptimizedBoolean(BlockScope currentScope, CodeStream codeStream, BranchLabel trueLabel, BranchLabel falseLabel, boolean valueRequired) {
    if (this.constant != Constant.NotAConstant && this.constant.typeID() == 5) {
      super.generateOptimizedBoolean(
          currentScope, 
          codeStream, 
          trueLabel, 
          falseLabel, 
          valueRequired);
      return;
    } 
    switch ((this.bits & 0xFC0) >> 6) {
      case 4:
        generateOptimizedLessThan(
            currentScope, 
            codeStream, 
            trueLabel, 
            falseLabel, 
            valueRequired);
        return;
      case 5:
        generateOptimizedLessThanOrEqual(
            currentScope, 
            codeStream, 
            trueLabel, 
            falseLabel, 
            valueRequired);
        return;
      case 6:
        generateOptimizedGreaterThan(
            currentScope, 
            codeStream, 
            trueLabel, 
            falseLabel, 
            valueRequired);
        return;
      case 7:
        generateOptimizedGreaterThanOrEqual(
            currentScope, 
            codeStream, 
            trueLabel, 
            falseLabel, 
            valueRequired);
        return;
      case 2:
        generateOptimizedLogicalAnd(
            currentScope, 
            codeStream, 
            trueLabel, 
            falseLabel, 
            valueRequired);
        return;
      case 3:
        generateOptimizedLogicalOr(
            currentScope, 
            codeStream, 
            trueLabel, 
            falseLabel, 
            valueRequired);
        return;
      case 8:
        generateOptimizedLogicalXor(
            currentScope, 
            codeStream, 
            trueLabel, 
            falseLabel, 
            valueRequired);
        return;
    } 
    super.generateOptimizedBoolean(
        currentScope, 
        codeStream, 
        trueLabel, 
        falseLabel, 
        valueRequired);
  }
  
  public void generateOptimizedGreaterThan(BlockScope currentScope, CodeStream codeStream, BranchLabel trueLabel, BranchLabel falseLabel, boolean valueRequired) {
    int promotedTypeID = (this.left.implicitConversion & 0xFF) >> 4;
    if (promotedTypeID == 10) {
      if (this.left.constant != Constant.NotAConstant && this.left.constant.intValue() == 0) {
        this.right.generateCode(currentScope, codeStream, valueRequired);
        if (valueRequired)
          if (falseLabel == null) {
            if (trueLabel != null)
              codeStream.iflt(trueLabel); 
          } else if (trueLabel == null) {
            codeStream.ifge(falseLabel);
          }  
        codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        return;
      } 
      if (this.right.constant != Constant.NotAConstant && this.right.constant.intValue() == 0) {
        this.left.generateCode(currentScope, codeStream, valueRequired);
        if (valueRequired)
          if (falseLabel == null) {
            if (trueLabel != null)
              codeStream.ifgt(trueLabel); 
          } else if (trueLabel == null) {
            codeStream.ifle(falseLabel);
          }  
        codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        return;
      } 
    } 
    this.left.generateCode(currentScope, codeStream, valueRequired);
    this.right.generateCode(currentScope, codeStream, valueRequired);
    if (valueRequired)
      if (falseLabel == null) {
        if (trueLabel != null) {
          switch (promotedTypeID) {
            case 10:
              codeStream.if_icmpgt(trueLabel);
              break;
            case 9:
              codeStream.fcmpl();
              codeStream.ifgt(trueLabel);
              break;
            case 7:
              codeStream.lcmp();
              codeStream.ifgt(trueLabel);
              break;
            case 8:
              codeStream.dcmpl();
              codeStream.ifgt(trueLabel);
              break;
          } 
          codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
          return;
        } 
      } else if (trueLabel == null) {
        switch (promotedTypeID) {
          case 10:
            codeStream.if_icmple(falseLabel);
            break;
          case 9:
            codeStream.fcmpl();
            codeStream.ifle(falseLabel);
            break;
          case 7:
            codeStream.lcmp();
            codeStream.ifle(falseLabel);
            break;
          case 8:
            codeStream.dcmpl();
            codeStream.ifle(falseLabel);
            break;
        } 
        codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        return;
      }  
  }
  
  public void generateOptimizedGreaterThanOrEqual(BlockScope currentScope, CodeStream codeStream, BranchLabel trueLabel, BranchLabel falseLabel, boolean valueRequired) {
    int promotedTypeID = (this.left.implicitConversion & 0xFF) >> 4;
    if (promotedTypeID == 10) {
      if (this.left.constant != Constant.NotAConstant && this.left.constant.intValue() == 0) {
        this.right.generateCode(currentScope, codeStream, valueRequired);
        if (valueRequired)
          if (falseLabel == null) {
            if (trueLabel != null)
              codeStream.ifle(trueLabel); 
          } else if (trueLabel == null) {
            codeStream.ifgt(falseLabel);
          }  
        codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        return;
      } 
      if (this.right.constant != Constant.NotAConstant && this.right.constant.intValue() == 0) {
        this.left.generateCode(currentScope, codeStream, valueRequired);
        if (valueRequired)
          if (falseLabel == null) {
            if (trueLabel != null)
              codeStream.ifge(trueLabel); 
          } else if (trueLabel == null) {
            codeStream.iflt(falseLabel);
          }  
        codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        return;
      } 
    } 
    this.left.generateCode(currentScope, codeStream, valueRequired);
    this.right.generateCode(currentScope, codeStream, valueRequired);
    if (valueRequired)
      if (falseLabel == null) {
        if (trueLabel != null) {
          switch (promotedTypeID) {
            case 10:
              codeStream.if_icmpge(trueLabel);
              break;
            case 9:
              codeStream.fcmpl();
              codeStream.ifge(trueLabel);
              break;
            case 7:
              codeStream.lcmp();
              codeStream.ifge(trueLabel);
              break;
            case 8:
              codeStream.dcmpl();
              codeStream.ifge(trueLabel);
              break;
          } 
          codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
          return;
        } 
      } else if (trueLabel == null) {
        switch (promotedTypeID) {
          case 10:
            codeStream.if_icmplt(falseLabel);
            break;
          case 9:
            codeStream.fcmpl();
            codeStream.iflt(falseLabel);
            break;
          case 7:
            codeStream.lcmp();
            codeStream.iflt(falseLabel);
            break;
          case 8:
            codeStream.dcmpl();
            codeStream.iflt(falseLabel);
            break;
        } 
        codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        return;
      }  
  }
  
  public void generateOptimizedLessThan(BlockScope currentScope, CodeStream codeStream, BranchLabel trueLabel, BranchLabel falseLabel, boolean valueRequired) {
    int promotedTypeID = (this.left.implicitConversion & 0xFF) >> 4;
    if (promotedTypeID == 10) {
      if (this.left.constant != Constant.NotAConstant && this.left.constant.intValue() == 0) {
        this.right.generateCode(currentScope, codeStream, valueRequired);
        if (valueRequired)
          if (falseLabel == null) {
            if (trueLabel != null)
              codeStream.ifgt(trueLabel); 
          } else if (trueLabel == null) {
            codeStream.ifle(falseLabel);
          }  
        codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        return;
      } 
      if (this.right.constant != Constant.NotAConstant && this.right.constant.intValue() == 0) {
        this.left.generateCode(currentScope, codeStream, valueRequired);
        if (valueRequired)
          if (falseLabel == null) {
            if (trueLabel != null)
              codeStream.iflt(trueLabel); 
          } else if (trueLabel == null) {
            codeStream.ifge(falseLabel);
          }  
        codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        return;
      } 
    } 
    this.left.generateCode(currentScope, codeStream, valueRequired);
    this.right.generateCode(currentScope, codeStream, valueRequired);
    if (valueRequired)
      if (falseLabel == null) {
        if (trueLabel != null) {
          switch (promotedTypeID) {
            case 10:
              codeStream.if_icmplt(trueLabel);
              break;
            case 9:
              codeStream.fcmpg();
              codeStream.iflt(trueLabel);
              break;
            case 7:
              codeStream.lcmp();
              codeStream.iflt(trueLabel);
              break;
            case 8:
              codeStream.dcmpg();
              codeStream.iflt(trueLabel);
              break;
          } 
          codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
          return;
        } 
      } else if (trueLabel == null) {
        switch (promotedTypeID) {
          case 10:
            codeStream.if_icmpge(falseLabel);
            break;
          case 9:
            codeStream.fcmpg();
            codeStream.ifge(falseLabel);
            break;
          case 7:
            codeStream.lcmp();
            codeStream.ifge(falseLabel);
            break;
          case 8:
            codeStream.dcmpg();
            codeStream.ifge(falseLabel);
            break;
        } 
        codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        return;
      }  
  }
  
  public void generateOptimizedLessThanOrEqual(BlockScope currentScope, CodeStream codeStream, BranchLabel trueLabel, BranchLabel falseLabel, boolean valueRequired) {
    int promotedTypeID = (this.left.implicitConversion & 0xFF) >> 4;
    if (promotedTypeID == 10) {
      if (this.left.constant != Constant.NotAConstant && this.left.constant.intValue() == 0) {
        this.right.generateCode(currentScope, codeStream, valueRequired);
        if (valueRequired)
          if (falseLabel == null) {
            if (trueLabel != null)
              codeStream.ifge(trueLabel); 
          } else if (trueLabel == null) {
            codeStream.iflt(falseLabel);
          }  
        codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        return;
      } 
      if (this.right.constant != Constant.NotAConstant && this.right.constant.intValue() == 0) {
        this.left.generateCode(currentScope, codeStream, valueRequired);
        if (valueRequired)
          if (falseLabel == null) {
            if (trueLabel != null)
              codeStream.ifle(trueLabel); 
          } else if (trueLabel == null) {
            codeStream.ifgt(falseLabel);
          }  
        codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        return;
      } 
    } 
    this.left.generateCode(currentScope, codeStream, valueRequired);
    this.right.generateCode(currentScope, codeStream, valueRequired);
    if (valueRequired)
      if (falseLabel == null) {
        if (trueLabel != null) {
          switch (promotedTypeID) {
            case 10:
              codeStream.if_icmple(trueLabel);
              break;
            case 9:
              codeStream.fcmpg();
              codeStream.ifle(trueLabel);
              break;
            case 7:
              codeStream.lcmp();
              codeStream.ifle(trueLabel);
              break;
            case 8:
              codeStream.dcmpg();
              codeStream.ifle(trueLabel);
              break;
          } 
          codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
          return;
        } 
      } else if (trueLabel == null) {
        switch (promotedTypeID) {
          case 10:
            codeStream.if_icmpgt(falseLabel);
            break;
          case 9:
            codeStream.fcmpg();
            codeStream.ifgt(falseLabel);
            break;
          case 7:
            codeStream.lcmp();
            codeStream.ifgt(falseLabel);
            break;
          case 8:
            codeStream.dcmpg();
            codeStream.ifgt(falseLabel);
            break;
        } 
        codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        return;
      }  
  }
  
  public void generateLogicalAnd(BlockScope currentScope, CodeStream codeStream, boolean valueRequired) {
    if ((this.left.implicitConversion & 0xF) == 5) {
      Constant condConst;
      if ((condConst = this.left.optimizedBooleanConstant()) != Constant.NotAConstant) {
        if (condConst.booleanValue()) {
          this.left.generateCode(currentScope, codeStream, false);
          this.right.generateCode(currentScope, codeStream, valueRequired);
        } else {
          this.left.generateCode(currentScope, codeStream, false);
          this.right.generateCode(currentScope, codeStream, false);
          if (valueRequired)
            codeStream.iconst_0(); 
          codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        } 
        return;
      } 
      if ((condConst = this.right.optimizedBooleanConstant()) != Constant.NotAConstant) {
        if (condConst.booleanValue()) {
          this.left.generateCode(currentScope, codeStream, valueRequired);
          this.right.generateCode(currentScope, codeStream, false);
        } else {
          this.left.generateCode(currentScope, codeStream, false);
          this.right.generateCode(currentScope, codeStream, false);
          if (valueRequired)
            codeStream.iconst_0(); 
          codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        } 
        return;
      } 
    } 
    this.left.generateCode(currentScope, codeStream, valueRequired);
    this.right.generateCode(currentScope, codeStream, valueRequired);
    if (valueRequired)
      codeStream.iand(); 
    codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
  }
  
  public void generateLogicalOr(BlockScope currentScope, CodeStream codeStream, boolean valueRequired) {
    if ((this.left.implicitConversion & 0xF) == 5) {
      Constant condConst;
      if ((condConst = this.left.optimizedBooleanConstant()) != Constant.NotAConstant) {
        if (condConst.booleanValue()) {
          this.left.generateCode(currentScope, codeStream, false);
          this.right.generateCode(currentScope, codeStream, false);
          if (valueRequired)
            codeStream.iconst_1(); 
          codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        } else {
          this.left.generateCode(currentScope, codeStream, false);
          this.right.generateCode(currentScope, codeStream, valueRequired);
        } 
        return;
      } 
      if ((condConst = this.right.optimizedBooleanConstant()) != Constant.NotAConstant) {
        if (condConst.booleanValue()) {
          this.left.generateCode(currentScope, codeStream, false);
          this.right.generateCode(currentScope, codeStream, false);
          if (valueRequired)
            codeStream.iconst_1(); 
          codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        } else {
          this.left.generateCode(currentScope, codeStream, valueRequired);
          this.right.generateCode(currentScope, codeStream, false);
        } 
        return;
      } 
    } 
    this.left.generateCode(currentScope, codeStream, valueRequired);
    this.right.generateCode(currentScope, codeStream, valueRequired);
    if (valueRequired)
      codeStream.ior(); 
    codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
  }
  
  public void generateLogicalXor(BlockScope currentScope, CodeStream codeStream, boolean valueRequired) {
    if ((this.left.implicitConversion & 0xF) == 5) {
      Constant condConst;
      if ((condConst = this.left.optimizedBooleanConstant()) != Constant.NotAConstant) {
        if (condConst.booleanValue()) {
          this.left.generateCode(currentScope, codeStream, false);
          if (valueRequired)
            codeStream.iconst_1(); 
          this.right.generateCode(currentScope, codeStream, valueRequired);
          if (valueRequired) {
            codeStream.ixor();
            codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
          } 
        } else {
          this.left.generateCode(currentScope, codeStream, false);
          this.right.generateCode(currentScope, codeStream, valueRequired);
        } 
        return;
      } 
      if ((condConst = this.right.optimizedBooleanConstant()) != Constant.NotAConstant) {
        if (condConst.booleanValue()) {
          this.left.generateCode(currentScope, codeStream, valueRequired);
          this.right.generateCode(currentScope, codeStream, false);
          if (valueRequired) {
            codeStream.iconst_1();
            codeStream.ixor();
            codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
          } 
        } else {
          this.left.generateCode(currentScope, codeStream, valueRequired);
          this.right.generateCode(currentScope, codeStream, false);
        } 
        return;
      } 
    } 
    this.left.generateCode(currentScope, codeStream, valueRequired);
    this.right.generateCode(currentScope, codeStream, valueRequired);
    if (valueRequired)
      codeStream.ixor(); 
    codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
  }
  
  public void generateOptimizedLogicalAnd(BlockScope currentScope, CodeStream codeStream, BranchLabel trueLabel, BranchLabel falseLabel, boolean valueRequired) {
    if ((this.left.implicitConversion & 0xF) == 5) {
      Constant condConst;
      if ((condConst = this.left.optimizedBooleanConstant()) != Constant.NotAConstant) {
        if (condConst.booleanValue()) {
          this.left.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              false);
          this.right.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              valueRequired);
        } else {
          this.left.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              false);
          this.right.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              false);
          if (valueRequired && 
            falseLabel != null)
            codeStream.goto_(falseLabel); 
          codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        } 
        return;
      } 
      if ((condConst = this.right.optimizedBooleanConstant()) != Constant.NotAConstant) {
        if (condConst.booleanValue()) {
          this.left.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              valueRequired);
          this.right.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              false);
        } else {
          BranchLabel internalTrueLabel = new BranchLabel(codeStream);
          this.left.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              internalTrueLabel, 
              falseLabel, 
              false);
          internalTrueLabel.place();
          this.right.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              false);
          if (valueRequired && 
            falseLabel != null)
            codeStream.goto_(falseLabel); 
          codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        } 
        return;
      } 
    } 
    this.left.generateCode(currentScope, codeStream, valueRequired);
    this.right.generateCode(currentScope, codeStream, valueRequired);
    if (valueRequired) {
      codeStream.iand();
      if (falseLabel == null) {
        if (trueLabel != null)
          codeStream.ifne(trueLabel); 
      } else if (trueLabel == null) {
        codeStream.ifeq(falseLabel);
      } 
    } 
    codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
  }
  
  public void generateOptimizedLogicalOr(BlockScope currentScope, CodeStream codeStream, BranchLabel trueLabel, BranchLabel falseLabel, boolean valueRequired) {
    if ((this.left.implicitConversion & 0xF) == 5) {
      Constant condConst;
      if ((condConst = this.left.optimizedBooleanConstant()) != Constant.NotAConstant) {
        if (condConst.booleanValue()) {
          this.left.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              false);
          BranchLabel internalFalseLabel = new BranchLabel(codeStream);
          this.right.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              internalFalseLabel, 
              false);
          internalFalseLabel.place();
          if (valueRequired && 
            trueLabel != null)
            codeStream.goto_(trueLabel); 
          codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        } else {
          this.left.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              false);
          this.right.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              valueRequired);
        } 
        return;
      } 
      if ((condConst = this.right.optimizedBooleanConstant()) != Constant.NotAConstant) {
        if (condConst.booleanValue()) {
          BranchLabel internalFalseLabel = new BranchLabel(codeStream);
          this.left.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              internalFalseLabel, 
              false);
          internalFalseLabel.place();
          this.right.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              false);
          if (valueRequired && 
            trueLabel != null)
            codeStream.goto_(trueLabel); 
          codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
        } else {
          this.left.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              valueRequired);
          this.right.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              false);
        } 
        return;
      } 
    } 
    this.left.generateCode(currentScope, codeStream, valueRequired);
    this.right.generateCode(currentScope, codeStream, valueRequired);
    if (valueRequired) {
      codeStream.ior();
      if (falseLabel == null) {
        if (trueLabel != null)
          codeStream.ifne(trueLabel); 
      } else if (trueLabel == null) {
        codeStream.ifeq(falseLabel);
      } 
    } 
    codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
  }
  
  public void generateOptimizedLogicalXor(BlockScope currentScope, CodeStream codeStream, BranchLabel trueLabel, BranchLabel falseLabel, boolean valueRequired) {
    if ((this.left.implicitConversion & 0xF) == 5) {
      Constant condConst;
      if ((condConst = this.left.optimizedBooleanConstant()) != Constant.NotAConstant) {
        if (condConst.booleanValue()) {
          this.left.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              false);
          this.right.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              falseLabel, 
              trueLabel, 
              valueRequired);
        } else {
          this.left.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              false);
          this.right.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              valueRequired);
        } 
        return;
      } 
      if ((condConst = this.right.optimizedBooleanConstant()) != Constant.NotAConstant) {
        if (condConst.booleanValue()) {
          this.left.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              falseLabel, 
              trueLabel, 
              valueRequired);
          this.right.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              false);
        } else {
          this.left.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              valueRequired);
          this.right.generateOptimizedBoolean(
              currentScope, 
              codeStream, 
              trueLabel, 
              falseLabel, 
              false);
        } 
        return;
      } 
    } 
    this.left.generateCode(currentScope, codeStream, valueRequired);
    this.right.generateCode(currentScope, codeStream, valueRequired);
    if (valueRequired) {
      codeStream.ixor();
      if (falseLabel == null) {
        if (trueLabel != null)
          codeStream.ifne(trueLabel); 
      } else if (trueLabel == null) {
        codeStream.ifeq(falseLabel);
      } 
    } 
    codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd);
  }
  
  public void generateOptimizedStringConcatenation(BlockScope blockScope, CodeStream codeStream, int typeID) {
    if ((this.bits & 0xFC0) >> 6 == 14 && (
      this.bits & 0xF) == 11) {
      if (this.constant != Constant.NotAConstant) {
        codeStream.generateConstant(this.constant, this.implicitConversion);
        codeStream.invokeStringConcatenationAppendForType(this.implicitConversion & 0xF);
      } else {
        int pc = codeStream.position;
        this.left.generateOptimizedStringConcatenation(
            blockScope, 
            codeStream, 
            this.left.implicitConversion & 0xF);
        codeStream.recordPositionsFrom(pc, this.left.sourceStart);
        pc = codeStream.position;
        this.right.generateOptimizedStringConcatenation(
            blockScope, 
            codeStream, 
            this.right.implicitConversion & 0xF);
        codeStream.recordPositionsFrom(pc, this.right.sourceStart);
      } 
    } else {
      super.generateOptimizedStringConcatenation(blockScope, codeStream, typeID);
    } 
  }
  
  public void generateOptimizedStringConcatenationCreation(BlockScope blockScope, CodeStream codeStream, int typeID) {
    if ((this.bits & 0xFC0) >> 6 == 14 && (
      this.bits & 0xF) == 11) {
      if (this.constant != Constant.NotAConstant) {
        codeStream.newStringContatenation();
        codeStream.dup();
        codeStream.ldc(this.constant.stringValue());
        codeStream.invokeStringConcatenationStringConstructor();
      } else {
        int pc = codeStream.position;
        this.left.generateOptimizedStringConcatenationCreation(
            blockScope, 
            codeStream, 
            this.left.implicitConversion & 0xF);
        codeStream.recordPositionsFrom(pc, this.left.sourceStart);
        pc = codeStream.position;
        this.right.generateOptimizedStringConcatenation(
            blockScope, 
            codeStream, 
            this.right.implicitConversion & 0xF);
        codeStream.recordPositionsFrom(pc, this.right.sourceStart);
      } 
    } else {
      super.generateOptimizedStringConcatenationCreation(blockScope, codeStream, typeID);
    } 
  }
  
  public boolean isCompactableOperation() {
    return true;
  }
  
  void nonRecursiveResolveTypeUpwards(BlockScope scope) {
    TypeBinding leftType = this.left.resolvedType;
    boolean rightIsCast;
    if (rightIsCast = this.right instanceof CastExpression)
      this.right.bits |= 0x20; 
    TypeBinding rightType = this.right.resolveType(scope);
    if (leftType == null || rightType == null) {
      this.constant = Constant.NotAConstant;
      return;
    } 
    int leftTypeID = leftType.id;
    int rightTypeID = rightType.id;
    boolean use15specifics = ((scope.compilerOptions()).sourceLevel >= 3211264L);
    if (use15specifics) {
      if (!leftType.isBaseType() && rightTypeID != 11 && rightTypeID != 12)
        leftTypeID = (scope.environment().computeBoxingType(leftType)).id; 
      if (!rightType.isBaseType() && leftTypeID != 11 && leftTypeID != 12)
        rightTypeID = (scope.environment().computeBoxingType(rightType)).id; 
    } 
    if (leftTypeID > 15 || 
      rightTypeID > 15)
      if (leftTypeID == 11) {
        rightTypeID = 1;
      } else if (rightTypeID == 11) {
        leftTypeID = 1;
      } else {
        this.constant = Constant.NotAConstant;
        scope.problemReporter().invalidOperator(this, leftType, rightType);
        return;
      }  
    if ((this.bits & 0xFC0) >> 6 == 14) {
      if (leftTypeID == 11) {
        this.left.computeConversion((Scope)scope, leftType, leftType);
        if (rightType.isArrayType() && TypeBinding.equalsEquals(((ArrayBinding)rightType).elementsType(), (TypeBinding)TypeBinding.CHAR))
          scope.problemReporter().signalNoImplicitStringConversionForCharArrayExpression(this.right); 
      } 
      if (rightTypeID == 11) {
        this.right.computeConversion((Scope)scope, rightType, rightType);
        if (leftType.isArrayType() && TypeBinding.equalsEquals(((ArrayBinding)leftType).elementsType(), (TypeBinding)TypeBinding.CHAR))
          scope.problemReporter().signalNoImplicitStringConversionForCharArrayExpression(this.left); 
      } 
    } 
    int operator = (this.bits & 0xFC0) >> 6;
    int operatorSignature = OperatorExpression.OperatorSignatures[operator][(leftTypeID << 4) + rightTypeID];
    this.left.computeConversion((Scope)scope, TypeBinding.wellKnownType((Scope)scope, operatorSignature >>> 16 & 0xF), leftType);
    this.right.computeConversion((Scope)scope, TypeBinding.wellKnownType((Scope)scope, operatorSignature >>> 8 & 0xF), rightType);
    this.bits |= operatorSignature & 0xF;
    switch (operatorSignature & 0xF) {
      case 5:
        this.resolvedType = (TypeBinding)TypeBinding.BOOLEAN;
        break;
      case 3:
        this.resolvedType = (TypeBinding)TypeBinding.BYTE;
        break;
      case 2:
        this.resolvedType = (TypeBinding)TypeBinding.CHAR;
        break;
      case 8:
        this.resolvedType = (TypeBinding)TypeBinding.DOUBLE;
        break;
      case 9:
        this.resolvedType = (TypeBinding)TypeBinding.FLOAT;
        break;
      case 10:
        this.resolvedType = (TypeBinding)TypeBinding.INT;
        break;
      case 7:
        this.resolvedType = (TypeBinding)TypeBinding.LONG;
        break;
      case 11:
        this.resolvedType = (TypeBinding)scope.getJavaLangString();
        break;
      default:
        this.constant = Constant.NotAConstant;
        scope.problemReporter().invalidOperator(this, leftType, rightType);
        return;
    } 
    boolean leftIsCast;
    if ((leftIsCast = this.left instanceof CastExpression) || 
      rightIsCast)
      CastExpression.checkNeedForArgumentCasts(scope, operator, operatorSignature, this.left, leftTypeID, leftIsCast, this.right, rightTypeID, rightIsCast); 
    computeConstant(scope, leftTypeID, rightTypeID);
  }
  
  public void optimizedBooleanConstant(int leftId, int operator, int rightId) {
    Constant cst;
    switch (operator) {
      case 2:
        if (leftId != 5 || rightId != 5)
          return; 
      case 0:
        if ((cst = this.left.optimizedBooleanConstant()) != Constant.NotAConstant) {
          if (!cst.booleanValue()) {
            this.optimizedBooleanConstant = cst;
            return;
          } 
          if ((cst = this.right.optimizedBooleanConstant()) != Constant.NotAConstant)
            this.optimizedBooleanConstant = cst; 
          return;
        } 
        if ((cst = this.right.optimizedBooleanConstant()) != Constant.NotAConstant && 
          !cst.booleanValue())
          this.optimizedBooleanConstant = cst; 
        return;
      case 3:
        if (leftId != 5 || rightId != 5)
          return; 
      case 1:
        if ((cst = this.left.optimizedBooleanConstant()) != Constant.NotAConstant) {
          if (cst.booleanValue()) {
            this.optimizedBooleanConstant = cst;
            return;
          } 
          if ((cst = this.right.optimizedBooleanConstant()) != Constant.NotAConstant)
            this.optimizedBooleanConstant = cst; 
          return;
        } 
        if ((cst = this.right.optimizedBooleanConstant()) != Constant.NotAConstant && 
          cst.booleanValue())
          this.optimizedBooleanConstant = cst; 
        break;
    } 
  }
  
  public StringBuffer printExpressionNoParenthesis(int indent, StringBuffer output) {
    this.left.printExpression(indent, output).append(' ').append(operatorToString()).append(' ');
    return this.right.printExpression(0, output);
  }
  
  public TypeBinding resolveType(BlockScope scope) {
    boolean leftIsCast;
    if (leftIsCast = this.left instanceof CastExpression)
      this.left.bits |= 0x20; 
    TypeBinding leftType = this.left.resolveType(scope);
    boolean rightIsCast;
    if (rightIsCast = this.right instanceof CastExpression)
      this.right.bits |= 0x20; 
    TypeBinding rightType = this.right.resolveType(scope);
    if (leftType == null || rightType == null) {
      this.constant = Constant.NotAConstant;
      return null;
    } 
    int leftTypeID = leftType.id;
    int rightTypeID = rightType.id;
    boolean use15specifics = ((scope.compilerOptions()).sourceLevel >= 3211264L);
    if (use15specifics) {
      if (!leftType.isBaseType() && rightTypeID != 11 && rightTypeID != 12)
        leftTypeID = (scope.environment().computeBoxingType(leftType)).id; 
      if (!rightType.isBaseType() && leftTypeID != 11 && leftTypeID != 12)
        rightTypeID = (scope.environment().computeBoxingType(rightType)).id; 
    } 
    if (leftTypeID > 15 || 
      rightTypeID > 15)
      if (leftTypeID == 11) {
        rightTypeID = 1;
      } else if (rightTypeID == 11) {
        leftTypeID = 1;
      } else {
        this.constant = Constant.NotAConstant;
        scope.problemReporter().invalidOperator(this, leftType, rightType);
        return null;
      }  
    if ((this.bits & 0xFC0) >> 6 == 14) {
      if (leftTypeID == 11) {
        this.left.computeConversion((Scope)scope, leftType, leftType);
        if (rightType.isArrayType() && TypeBinding.equalsEquals(((ArrayBinding)rightType).elementsType(), (TypeBinding)TypeBinding.CHAR))
          scope.problemReporter().signalNoImplicitStringConversionForCharArrayExpression(this.right); 
      } 
      if (rightTypeID == 11) {
        this.right.computeConversion((Scope)scope, rightType, rightType);
        if (leftType.isArrayType() && TypeBinding.equalsEquals(((ArrayBinding)leftType).elementsType(), (TypeBinding)TypeBinding.CHAR))
          scope.problemReporter().signalNoImplicitStringConversionForCharArrayExpression(this.left); 
      } 
    } 
    int operator = (this.bits & 0xFC0) >> 6;
    int operatorSignature = OperatorExpression.OperatorSignatures[operator][(leftTypeID << 4) + rightTypeID];
    this.left.computeConversion((Scope)scope, TypeBinding.wellKnownType((Scope)scope, operatorSignature >>> 16 & 0xF), leftType);
    this.right.computeConversion((Scope)scope, TypeBinding.wellKnownType((Scope)scope, operatorSignature >>> 8 & 0xF), rightType);
    this.bits |= operatorSignature & 0xF;
    switch (operatorSignature & 0xF) {
      case 5:
        this.resolvedType = (TypeBinding)TypeBinding.BOOLEAN;
        break;
      case 3:
        this.resolvedType = (TypeBinding)TypeBinding.BYTE;
        break;
      case 2:
        this.resolvedType = (TypeBinding)TypeBinding.CHAR;
        break;
      case 8:
        this.resolvedType = (TypeBinding)TypeBinding.DOUBLE;
        break;
      case 9:
        this.resolvedType = (TypeBinding)TypeBinding.FLOAT;
        break;
      case 10:
        this.resolvedType = (TypeBinding)TypeBinding.INT;
        break;
      case 7:
        this.resolvedType = (TypeBinding)TypeBinding.LONG;
        break;
      case 11:
        this.resolvedType = (TypeBinding)scope.getJavaLangString();
        break;
      default:
        this.constant = Constant.NotAConstant;
        scope.problemReporter().invalidOperator(this, leftType, rightType);
        return null;
    } 
    if (leftIsCast || rightIsCast)
      CastExpression.checkNeedForArgumentCasts(scope, operator, operatorSignature, this.left, leftTypeID, leftIsCast, this.right, rightTypeID, rightIsCast); 
    computeConstant(scope, leftTypeID, rightTypeID);
    return this.resolvedType;
  }
  
  public void traverse(ASTVisitor visitor, BlockScope scope) {
    if (visitor.visit(this, scope)) {
      this.left.traverse(visitor, scope);
      this.right.traverse(visitor, scope);
    } 
    visitor.endVisit(this, scope);
  }
}
